/*
 * Copyright (C) 2007 Dell Inc.
 * Chieng-Hwa Lin <chieng-hwa_lin@dell.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/*
 *  History:
 */

#include <linux/version.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/timer.h>
#include <linux/mtio.h>
#include <linux/ioctl.h>
#include <linux/fcntl.h>
// 2.6.24
#include <linux/interrupt.h>
#include <linux/poll.h>
#include <linux/delay.h>
#include <linux/wait.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,2)
#include <linux/vmalloc.h>
#endif

#include "./include/gpiotypes.h"
#include "./include/gpio_struct.h"
#include "./include/gpioiocts.h"
#include <cmc_gpio.h>

extern GPIO_DEVICE_EXTENSION gpio_card;
extern spinlock_t    ControlLock_bhlock;
extern spinlock_t    ControlLock_irqlock;
extern wait_queue_head_t FanInterruptSemaphore;
extern wait_queue_head_t TempInterruptSemaphore;
extern wait_queue_head_t ShftgInterruptSemaphore;  // Sempahore for shifty interrupt
extern wait_queue_head_t BladeInterruptSemaphore;  // Sempahore for blade interrupt
extern wait_queue_head_t IOMInterruptSemaphore;  // Sempahore for IOM interrupt
extern wait_queue_head_t ECCInterruptSemaphore;  // Sempahore for IOM interrupt
extern wait_queue_head_t XilinxRecoverySemaphore;  // Sempahore for I2C1 recovery
extern wait_queue_head_t XilinxDisableSemaphore;  // Sempahore for I2C1 disable
extern wait_queue_head_t PSUInterruptSemaphore;  // Sempahore for PSU interrupt
extern wait_queue_head_t CPInterruptSemaphore;  // Sempahore for CP interrupt
extern wait_queue_head_t KVMInterruptSemaphore;  // Sempahore for CP interrupt


//--------------------------------------------------------------------
// Local definitions
//--------------------------------------------------------------------

int ECC_event_present;
int xilinx_recovery_event_present;
int xilinx_disable_event_present;
int driver_exited = 0;

struct intr_status_t *current_blade_intr_status;

extern struct intr_status_t blade_intr_queue;
extern struct intr_status_t fan_intr_queue;
extern struct intr_status_t iom_intr_queue;
extern struct intr_status_t psu_intr_queue;
extern struct intr_status_t kvm_intr_queue;
extern struct intr_status_t shifty_intr_queue;
extern struct intr_status_t cp_intr_queue;

//--------------------------------------------------------------------
// Prototypes
//--------------------------------------------------------------------
int ProcessInterruptQueue(struct intr_status_t *intr_queue,
		spinlock_t *irqlock,
		wait_queue_head_t *semp,
		u32 *GPIOReg);
void CMCGpioSetBits32(int base_address, int offset, u32 mask, u32 val);
int  
gpio_i2c_dbg(
    const char *myformat,
    ...
    );
void gpio_spin_lock(PGPIO_DEVICE_EXTENSION pGpioCard);
void gpio_spin_unlock(PGPIO_DEVICE_EXTENSION pGpioCard);

u32 QueueISRStatus(struct intr_status_t *intr_queue, u32 intr_status,
			wait_queue_head_t *semaphore);

//--------------------------------------------------------------------
// Code start
//--------------------------------------------------------------------
static u32
paddr_to_vaddr(u32 base_address, u32 offset)
{
	u32 paddr;
	u32 pbase;
	u32 vaddr;

	paddr = base_address + offset;
	if (paddr >= CMC_GPIO_BASE_ADDRESS &&
			paddr < CMC_GPIO_BASE_ADDRESS+8192) {
		pbase = CMC_GPIO_BASE_ADDRESS;
		vaddr = gpio_card.GPIOMappedBaseAddress;
	} else if (paddr >= CMC_PPC_IIC0_BASE_ADDRESS &&
			paddr < CMC_PPC_IIC0_BASE_ADDRESS+32) {
		pbase = CMC_PPC_IIC0_BASE_ADDRESS;
		vaddr = gpio_card.IIC0MappedBaseAddress;
	} else if (paddr >= CMC_PPC_IIC1_BASE_ADDRESS &&
			paddr < CMC_PPC_IIC1_BASE_ADDRESS+32) {
		pbase = CMC_PPC_IIC1_BASE_ADDRESS;
		vaddr = gpio_card.IIC1MappedBaseAddress;
	} else if (paddr >= CMC_PPC_GPIO_0_BASE_ADDRESS &&
			paddr < CMC_PPC_GPIO_0_BASE_ADDRESS+128) {
		pbase = CMC_PPC_GPIO_0_BASE_ADDRESS;
	    	vaddr = gpio_card.GPIO0MappedBaseAddress;
	} else if (paddr >= CMC_PPC_GPIO_0_BASE_ADDRESS_1 &&
			paddr < CMC_PPC_GPIO_0_BASE_ADDRESS_1+128) {
		pbase = CMC_PPC_GPIO_0_BASE_ADDRESS_1;
	      	vaddr = gpio_card.GPIO1MappedBaseAddress;
	} else
		panic("CMCGpio[Read|Write]Reg[32]: bad address");
	vaddr += paddr - pbase;
	return(vaddr);
}

//---------------------------------------------------------------------
// CMCGpioReadReg - read the GPIO register
//---------------------------------------------------------------------
void CMCGpioReadReg(int base_address, int offset, int *reg_result)
{
	u32 VirtAddress = paddr_to_vaddr(base_address, offset);

	// Do GPIO stuff
	*reg_result = *(u16 *)VirtAddress;
} // CMCGpioReadReg

//---------------------------------------------------------------------
// CMCGpioWriteReg - write the GPIO register
//---------------------------------------------------------------------
void CMCGpioWriteReg(int base_address, int offset, u16 reg_input)
{
	u32 temp1 = base_address + offset;
	u32 VirtAddress = paddr_to_vaddr(base_address, offset);

	/*
	 * Manage /proc filesystem.
	 */
	if (temp1 == (CMC_FPGA_BASE_ADDRESS+CMC_FPGA_MODE_0_REG_OFFSET)) {
		if (reg_input&MODE_VENDOR_SOFT)
			create_cmc_file(SOFT_VENDOR_MODE);
		else
			remove_cmc_file(SOFT_VENDOR_MODE);
	}
	*(u16 *)VirtAddress  = reg_input;
} // CMCGpioWriteReg

//---------------------------------------------------------------------
// CMCGpioSetBits - write the GPIO register
//---------------------------------------------------------------------
static DEFINE_SPINLOCK(gpiobits_irqlock);
void CMCGpioSetBits(int base_address, int offset, u16 mask, u16 val)
{
	u32 temp1 = base_address + offset;
	u32 VirtAddress = paddr_to_vaddr(base_address, offset);
	unsigned long flags;
	u16 data;

	/*
	 * Manage /proc filesystem.
	 */
	if (temp1 == (CMC_FPGA_BASE_ADDRESS + CMC_FPGA_MODE_0_REG_OFFSET) &&
		(mask&MODE_VENDOR_SOFT))
	{
		if (val&MODE_VENDOR_SOFT)
			create_cmc_file(SOFT_VENDOR_MODE);
		else
			remove_cmc_file(SOFT_VENDOR_MODE);
	}
	spin_lock_irqsave(&gpiobits_irqlock, flags);
	data = *(u16 *)VirtAddress & ~mask;
	*(u16 *)VirtAddress = data | (mask&val);
	spin_unlock_irqrestore(&gpiobits_irqlock, flags);
} // CMCGpioSetBits

//---------------------------------------------------------------------
// CMCGpioClrBits - write the GPIO register
//---------------------------------------------------------------------
void CMCGpioClrBits(int base_address, int offset, u16 mask)
{
	CMCGpioSetBits(base_address, offset, mask, 0);
} // CMCGpioClrBits

//---------------------------------------------------------------------
// CMCGpioReadReg32 - read the GPIO register
//---------------------------------------------------------------------
void CMCGpioReadReg32(u32 base_address, u32 offset, u32 *reg_result)
{
	u32 VirtAddress = paddr_to_vaddr(base_address, offset);

	// Do GPIO stuff
	*reg_result = *(u32 *)VirtAddress;
	//printk("CMCGpioReadReg32, Result: 0x%x\n", *reg_result);
} // CMCGpioReadReg32

//---------------------------------------------------------------------
// CMCGpioWriteReg32 - write the GPIO register
//---------------------------------------------------------------------
void CMCGpioWriteReg32(int base_address, int offset, u32 reg_input)
{
	u32 temp1 = base_address + offset;
	u32 VirtAddress;

	/*
	 * UART and NIC-releated GPIOs are managed by their respective drivers.
	 */
	if (temp1 == 0xef600b00)
	{
		CMCGpioSetBits32(base_address, offset,
			~(GPIO_PPC_UART0_DTR|GPIO_UART_CMC_IOM_DTR_B),
			reg_input);
		return;
	}
	if (temp1 == 0xef600c00)
	{
		CMCGpioSetBits32(base_address, offset,
			~(GPIO_PPC_MICREL_EN),
			reg_input);
		return;
	}
	VirtAddress = paddr_to_vaddr(base_address, offset);

	*(u32 *)VirtAddress = reg_input;
} // CMCGpioWriteReg32

//---------------------------------------------------------------------
// CMCGpioSetBits32 - write the GPIO register
//---------------------------------------------------------------------
void CMCGpioSetBits32(int base_address, int offset, u32 mask, u32 val)
{
	u32 VirtAddress = paddr_to_vaddr(base_address, offset);
	unsigned long flags;
	u32 data;

	spin_lock_irqsave(&gpiobits_irqlock, flags);
	data = *(u32 *)VirtAddress & ~mask;
	*(u32 *)VirtAddress = data | (mask&val);
	spin_unlock_irqrestore(&gpiobits_irqlock, flags);
} // CMCGpioSetBits32

//---------------------------------------------------------------------
// CMCGpioClrBits32 - write the GPIO register
//---------------------------------------------------------------------
void CMCGpioClrBits32(int base_address, int offset, u32 mask)
{
	CMCGpioSetBits32(base_address, offset, mask, 0);
} // CMCGpioClrBits32

//---------------------------------------------------------------------
// GpioDeviceIoControl - Process all IOCTLS coming from
// main driver (dcgpio.c)
//---------------------------------------------------------------------

int GpioDeviceIoControl(struct inode * inode,
	struct file * filp,
	PGPIO_DEVICE_EXTENSION pGpioCard,
	unsigned int cmd_in,
	unsigned long arg,
	int *lpBytesReturned)
{
	GpioCmdArg *pGpioCmdArgBuf;

	u32 BytesReturned;
	u32 result;
	u32 IoCmdStatus;
	u32 GPIOReg;
	u32 temp;
	u32 status;

	temp = sizeof(GpioCmdArg);
	pGpioCmdArgBuf = (GpioCmdArg *)kmalloc(temp, GFP_KERNEL);	
    
	if (pGpioCmdArgBuf == NULL)
	{
		return -1;
	}

	if (_IOC_TYPE(cmd_in) != GPIO_IOC_MAGIC) 
	{	
		
			kfree (pGpioCmdArgBuf);
			return -EINVAL;
	}

	// zero out the return length before returning status
	*lpBytesReturned = 0;


		// setup default values
	IoCmdStatus = 0;

	// chl, we are different form Netware/NT here.
	// We need to copy the data from user's space to
	// kernel space before we can do anything.
	// Also, we only have one pointer for all arguments, so
	// we need to check the cmd type in order to get all 
	// arguments from our structures of union.

    switch(cmd_in)
    {
	case IOCTL_GPIO_SETBITS:
	{
	    // User space pointer
            GpioCmdArg *pUserSpaceGPIOCmdIoctlArg;
	    // Command request
            GpioCmdIoctlReq *pGpioGPIOCmdIoctlBuf;
	    // RespBuf pointer
	    GpioCmdIoctlReq *pGpioGPIORespIoctlBuf;


	    pUserSpaceGPIOCmdIoctlArg = (GpioCmdArg *) arg;

	    result = copy_from_user((char *)pGpioCmdArgBuf, 
		(char *)pUserSpaceGPIOCmdIoctlArg, sizeof(GpioCmdArg));
            if (result) 
	    {
		IoCmdStatus = result;
		break;
	    }

	    // CmdBuf pointer
	    pGpioGPIOCmdIoctlBuf = &pGpioCmdArgBuf->CmdBuf;
	
	    // RespBuf pointer
	    pGpioGPIORespIoctlBuf = &pGpioCmdArgBuf->RespBuf;

	    memmove((void *)pGpioGPIORespIoctlBuf, 
			(void *)pGpioGPIOCmdIoctlBuf, 
			sizeof(GpioCmdIoctlReq));

	    CMCGpioSetBits(pGpioCmdArgBuf->CmdBuf.BaseAddress,
		    pGpioCmdArgBuf->CmdBuf.Offset,
		    pGpioCmdArgBuf->CmdBuf.Mask,
		    pGpioCmdArgBuf->CmdBuf.Value);

	    // set status of IOCTL
	    pGpioGPIORespIoctlBuf->Status = IOCTL_STATUS_SUCCESS;
	    BytesReturned = 0;
	    IoCmdStatus = GPIO_STATUS_CMD_SUCCESS; // was executed
	    IoCmdStatus = copy_to_user((char *)&pUserSpaceGPIOCmdIoctlArg->RespBuf,
		    (char *)pGpioGPIORespIoctlBuf, sizeof(GpioCmdIoctlReq));
	    put_user(IoCmdStatus,
		    &pUserSpaceGPIOCmdIoctlArg->RespBuf.Status);
	    put_user(BytesReturned,
		  &pUserSpaceGPIOCmdIoctlArg->ReturnedLen);
	}
	break;
	case IOCTL_GPIO_SETBITS32:
	{
	    // User space pointer
            GpioCmdArg *pUserSpaceGPIOCmdIoctlArg;
	    // Command request
            GpioCmdIoctlReq *pGpioGPIOCmdIoctlBuf;
	    // RespBuf pointer
	    GpioCmdIoctlReq *pGpioGPIORespIoctlBuf;


	    pUserSpaceGPIOCmdIoctlArg = (GpioCmdArg *) arg;

	    result = copy_from_user((char *)pGpioCmdArgBuf, 
		(char *)pUserSpaceGPIOCmdIoctlArg, sizeof(GpioCmdArg));
            if (result) 
	    {
		IoCmdStatus = result;
		break;
	    }

	    // CmdBuf pointer
	    pGpioGPIOCmdIoctlBuf = &pGpioCmdArgBuf->CmdBuf;
	
	    // RespBuf pointer
	    pGpioGPIORespIoctlBuf = &pGpioCmdArgBuf->RespBuf;

	    memmove((void *)pGpioGPIORespIoctlBuf, 
			(void *)pGpioGPIOCmdIoctlBuf, 
			sizeof(GpioCmdIoctlReq));

	    CMCGpioSetBits32(pGpioCmdArgBuf->CmdBuf.BaseAddress,
		    pGpioCmdArgBuf->CmdBuf.Offset,
		    pGpioCmdArgBuf->CmdBuf.Mask,
		    pGpioCmdArgBuf->CmdBuf.Value);

	    // set status of IOCTL
	    pGpioGPIORespIoctlBuf->Status = IOCTL_STATUS_SUCCESS;
	    BytesReturned = 0;
	    IoCmdStatus = GPIO_STATUS_CMD_SUCCESS; // was executed
	    IoCmdStatus = copy_to_user((char *)&pUserSpaceGPIOCmdIoctlArg->RespBuf,
		    (char *)pGpioGPIORespIoctlBuf, sizeof(GpioCmdIoctlReq));
	    put_user(IoCmdStatus,
		    &pUserSpaceGPIOCmdIoctlArg->RespBuf.Status);
	    put_user(BytesReturned,
		  &pUserSpaceGPIOCmdIoctlArg->ReturnedLen);
	}
	break;
	case IOCTL_GPIO_CLRBITS:
	{
	    // User space pointer
            GpioCmdArg *pUserSpaceGPIOCmdIoctlArg;
	    // Command request
            GpioCmdIoctlReq *pGpioGPIOCmdIoctlBuf;
	    // RespBuf pointer
	    GpioCmdIoctlReq *pGpioGPIORespIoctlBuf;


	    pUserSpaceGPIOCmdIoctlArg = (GpioCmdArg *) arg;

	    result = copy_from_user((char *)pGpioCmdArgBuf, 
		(char *)pUserSpaceGPIOCmdIoctlArg, sizeof(GpioCmdArg));
            if (result) 
	    {
		IoCmdStatus = result;
		break;
	    }

	    // CmdBuf pointer
	    pGpioGPIOCmdIoctlBuf = &pGpioCmdArgBuf->CmdBuf;
	
	    // RespBuf pointer
	    pGpioGPIORespIoctlBuf = &pGpioCmdArgBuf->RespBuf;

	    memmove((void *)pGpioGPIORespIoctlBuf, 
			(void *)pGpioGPIOCmdIoctlBuf, 
			sizeof(GpioCmdIoctlReq));

	    CMCGpioClrBits(pGpioCmdArgBuf->CmdBuf.BaseAddress,
		    pGpioCmdArgBuf->CmdBuf.Offset,
		    pGpioCmdArgBuf->CmdBuf.Mask);

	    // set status of IOCTL
	    pGpioGPIORespIoctlBuf->Status = IOCTL_STATUS_SUCCESS;
	    BytesReturned = 0;
	    IoCmdStatus = GPIO_STATUS_CMD_SUCCESS; // was executed
	    IoCmdStatus = copy_to_user((char *)&pUserSpaceGPIOCmdIoctlArg->RespBuf,
		    (char *)pGpioGPIORespIoctlBuf, sizeof(GpioCmdIoctlReq));
	    put_user(IoCmdStatus,
		    &pUserSpaceGPIOCmdIoctlArg->RespBuf.Status);
	    put_user(BytesReturned,
		  &pUserSpaceGPIOCmdIoctlArg->ReturnedLen);
	}
	break;
	case IOCTL_GPIO_CLRBITS32:
	{
	    // User space pointer
            GpioCmdArg *pUserSpaceGPIOCmdIoctlArg;
	    // Command request
            GpioCmdIoctlReq *pGpioGPIOCmdIoctlBuf;
	    // RespBuf pointer
	    GpioCmdIoctlReq *pGpioGPIORespIoctlBuf;


	    pUserSpaceGPIOCmdIoctlArg = (GpioCmdArg *) arg;

	    result = copy_from_user((char *)pGpioCmdArgBuf, 
		(char *)pUserSpaceGPIOCmdIoctlArg, sizeof(GpioCmdArg));
            if (result) 
	    {
		IoCmdStatus = result;
		break;
	    }

	    // CmdBuf pointer
	    pGpioGPIOCmdIoctlBuf = &pGpioCmdArgBuf->CmdBuf;
	
	    // RespBuf pointer
	    pGpioGPIORespIoctlBuf = &pGpioCmdArgBuf->RespBuf;

	    memmove((void *)pGpioGPIORespIoctlBuf, 
			(void *)pGpioGPIOCmdIoctlBuf, 
			sizeof(GpioCmdIoctlReq));

	    CMCGpioClrBits32(pGpioCmdArgBuf->CmdBuf.BaseAddress,
		    pGpioCmdArgBuf->CmdBuf.Offset,
		    pGpioCmdArgBuf->CmdBuf.Mask);

	    // set status of IOCTL
	    pGpioGPIORespIoctlBuf->Status = IOCTL_STATUS_SUCCESS;
	    BytesReturned = 0;
	    IoCmdStatus = GPIO_STATUS_CMD_SUCCESS; // was executed
	    IoCmdStatus = copy_to_user((char *)&pUserSpaceGPIOCmdIoctlArg->RespBuf,
		    (char *)pGpioGPIORespIoctlBuf, sizeof(GpioCmdIoctlReq));
	    put_user(IoCmdStatus,
		    &pUserSpaceGPIOCmdIoctlArg->RespBuf.Status);
	    put_user(BytesReturned,
		  &pUserSpaceGPIOCmdIoctlArg->ReturnedLen);
	}
	break;
	case IOCTL_GPIO_WRITE_CMD:
	{
	    // User space pointer
            GpioCmdArg *pUserSpaceGPIOCmdIoctlArg;
	    // Command request
            GpioCmdIoctlReq *pGpioGPIOCmdIoctlBuf;
	    // RespBuf pointer
	    GpioCmdIoctlReq *pGpioGPIORespIoctlBuf;


	    pUserSpaceGPIOCmdIoctlArg = (GpioCmdArg *) arg;

	    result = copy_from_user((char *)pGpioCmdArgBuf, 
		(char *)pUserSpaceGPIOCmdIoctlArg, sizeof(GpioCmdArg));
            if (result) 
	    {
		IoCmdStatus = result;
		break;
	    }

	    // CmdBuf pointer
	    pGpioGPIOCmdIoctlBuf = &pGpioCmdArgBuf->CmdBuf;
	
	    // RespBuf pointer
	    pGpioGPIORespIoctlBuf = &pGpioCmdArgBuf->RespBuf;

	    memmove((void *)pGpioGPIORespIoctlBuf, 
			(void *)pGpioGPIOCmdIoctlBuf, 
			sizeof(GpioCmdIoctlReq));

		CMCGpioWriteReg(pGpioCmdArgBuf->CmdBuf.BaseAddress,
			pGpioCmdArgBuf->CmdBuf.Offset,
			pGpioCmdArgBuf->CmdBuf.Value);

		// set status of IOCTL
		pGpioGPIORespIoctlBuf->Status = IOCTL_STATUS_SUCCESS;
		BytesReturned = 0;
		IoCmdStatus = GPIO_STATUS_CMD_SUCCESS; // was executed
		IoCmdStatus = copy_to_user((char *)&pUserSpaceGPIOCmdIoctlArg->RespBuf,
		 	(char *)pGpioGPIORespIoctlBuf, sizeof(GpioCmdIoctlReq));
	    	put_user(IoCmdStatus,
            	  	&pUserSpaceGPIOCmdIoctlArg->RespBuf.Status);
		put_user(BytesReturned,
           	      &pUserSpaceGPIOCmdIoctlArg->ReturnedLen);
	}
	break;

	case IOCTL_GPIO_WRITE_CMD_32:
	{
	    // User space pointer
            GpioCmdArg *pUserSpaceGPIOCmdIoctlArg;
	    // Command request
            GpioCmdIoctlReq *pGpioGPIOCmdIoctlBuf;
	    // RespBuf pointer
	    GpioCmdIoctlReq *pGpioGPIORespIoctlBuf;


	    pUserSpaceGPIOCmdIoctlArg = (GpioCmdArg *) arg;

	    result = copy_from_user((char *)pGpioCmdArgBuf, 
		(char *)pUserSpaceGPIOCmdIoctlArg, sizeof(GpioCmdArg));
            if (result) 
	    {
		IoCmdStatus = result;
		break;
	    }

	    // CmdBuf pointer
	    pGpioGPIOCmdIoctlBuf = &pGpioCmdArgBuf->CmdBuf;
	
	    // RespBuf pointer
	    pGpioGPIORespIoctlBuf = &pGpioCmdArgBuf->RespBuf;

	    memmove((void *)pGpioGPIORespIoctlBuf, 
			(void *)pGpioGPIOCmdIoctlBuf, 
			sizeof(GpioCmdIoctlReq));

		CMCGpioWriteReg32(pGpioCmdArgBuf->CmdBuf.BaseAddress,
			pGpioCmdArgBuf->CmdBuf.Offset,
			pGpioCmdArgBuf->CmdBuf.Value);

		// set status of IOCTL
		pGpioGPIORespIoctlBuf->Status = IOCTL_STATUS_SUCCESS;
		BytesReturned = 0;
		IoCmdStatus = GPIO_STATUS_CMD_SUCCESS; // was executed
		IoCmdStatus = copy_to_user((char *)&pUserSpaceGPIOCmdIoctlArg->RespBuf,
		 	(char *)pGpioGPIORespIoctlBuf, sizeof(GpioCmdIoctlReq));
	    	put_user(IoCmdStatus,
            	  	&pUserSpaceGPIOCmdIoctlArg->RespBuf.Status);
		put_user(BytesReturned,
           	      &pUserSpaceGPIOCmdIoctlArg->ReturnedLen);
	}
	break;

	case IOCTL_GPIO_READ_CMD:
	{
	    // User space pointer
            GpioCmdArg *pUserSpaceGPIOCmdIoctlArg;
	    // Command request
            GpioCmdIoctlReq *pGpioGPIOCmdIoctlBuf;
	    // RespBuf pointer
	    GpioCmdIoctlReq *pGpioGPIORespIoctlBuf;

	    pUserSpaceGPIOCmdIoctlArg = (GpioCmdArg *) arg;

	    result = copy_from_user((char *)pGpioCmdArgBuf, 
		(char *)pUserSpaceGPIOCmdIoctlArg, sizeof(GpioCmdArg));
            if (result) 
	    {
		IoCmdStatus = result;
		break;
	    }

	    // CmdBuf pointer
	    pGpioGPIOCmdIoctlBuf = &pGpioCmdArgBuf->CmdBuf;
	
	    // RespBuf pointer
	    pGpioGPIORespIoctlBuf = &pGpioCmdArgBuf->RespBuf;

	    memmove((void *)pGpioGPIORespIoctlBuf, 
			(void *)pGpioGPIOCmdIoctlBuf, 
			sizeof(GpioCmdIoctlReq));

		CMCGpioReadReg(pGpioCmdArgBuf->CmdBuf.BaseAddress,
			pGpioCmdArgBuf->CmdBuf.Offset,
			&pGpioGPIORespIoctlBuf->Result);

		// set status of IOCTL
		pGpioGPIORespIoctlBuf->Status = IOCTL_STATUS_SUCCESS;
		BytesReturned = 0;
		IoCmdStatus = GPIO_STATUS_CMD_SUCCESS; // was executed
		IoCmdStatus = copy_to_user((char *)&pUserSpaceGPIOCmdIoctlArg->RespBuf,
		 	(char *)pGpioGPIORespIoctlBuf, sizeof(GpioCmdIoctlReq));
	    	put_user(IoCmdStatus,
            	  	&pUserSpaceGPIOCmdIoctlArg->RespBuf.Status);
		put_user(BytesReturned,
           	      &pUserSpaceGPIOCmdIoctlArg->ReturnedLen);
	}

	break;

	case IOCTL_GPIO_READ_CMD_32:
	{
	    // User space pointer
            GpioCmdArg *pUserSpaceGPIOCmdIoctlArg;
	    // Command request
            GpioCmdIoctlReq *pGpioGPIOCmdIoctlBuf;
	    // RespBuf pointer
	    GpioCmdIoctlReq *pGpioGPIORespIoctlBuf;

	    pUserSpaceGPIOCmdIoctlArg = (GpioCmdArg *) arg;

	    result = copy_from_user((char *)pGpioCmdArgBuf, 
		(char *)pUserSpaceGPIOCmdIoctlArg, sizeof(GpioCmdArg));
            if (result) 
	    {
		IoCmdStatus = result;
		break;
	    }

	    // CmdBuf pointer
	    pGpioGPIOCmdIoctlBuf = &pGpioCmdArgBuf->CmdBuf;
	
	    // RespBuf pointer
	    pGpioGPIORespIoctlBuf = &pGpioCmdArgBuf->RespBuf;

	    memmove((void *)pGpioGPIORespIoctlBuf, 
			(void *)pGpioGPIOCmdIoctlBuf, 
			sizeof(GpioCmdIoctlReq));

		CMCGpioReadReg32(pGpioCmdArgBuf->CmdBuf.BaseAddress,
			pGpioCmdArgBuf->CmdBuf.Offset,
			&pGpioGPIORespIoctlBuf->Result);

		// set status of IOCTL
		pGpioGPIORespIoctlBuf->Status = IOCTL_STATUS_SUCCESS;
		BytesReturned = 0;
		IoCmdStatus = GPIO_STATUS_CMD_SUCCESS; // was executed
		IoCmdStatus = copy_to_user((char *)&pUserSpaceGPIOCmdIoctlArg->RespBuf,
		 	(char *)pGpioGPIORespIoctlBuf, sizeof(GpioCmdIoctlReq));
	    	put_user(IoCmdStatus,
            	  	&pUserSpaceGPIOCmdIoctlArg->RespBuf.Status);
		put_user(BytesReturned,
           	      &pUserSpaceGPIOCmdIoctlArg->ReturnedLen);
	}

	break;

	case IOCTL_WAIT_FOR_FAN_INTR:
	{
	    // User space pointer
            GpioCmdArg *pUserSpaceGPIOCmdIoctlArg;
	    // Command request
            GpioCmdIoctlReq *pGpioGPIOCmdIoctlBuf;
	    // RespBuf pointer
	    GpioCmdIoctlReq *pGpioGPIORespIoctlBuf;

	    pUserSpaceGPIOCmdIoctlArg = (GpioCmdArg *) arg;

	    result = copy_from_user((char *)pGpioCmdArgBuf, 
		(char *)pUserSpaceGPIOCmdIoctlArg, sizeof(GpioCmdArg));
            if (result) 
	    {
		IoCmdStatus = result;
		break;
	    }

	    // CmdBuf pointer
	    pGpioGPIOCmdIoctlBuf = &pGpioCmdArgBuf->CmdBuf;
	
	    // RespBuf pointer
	    pGpioGPIORespIoctlBuf = &pGpioCmdArgBuf->RespBuf;

	    memmove((void *)pGpioGPIORespIoctlBuf, 
			(void *)pGpioGPIOCmdIoctlBuf, 
			sizeof(GpioCmdIoctlReq));
	    if ((IoCmdStatus = ProcessInterruptQueue(&fan_intr_queue,
			&pGpioCard->fan_controlLock_irqlock,
			&FanInterruptSemaphore,
			&GPIOReg)))
		goto get_out;

	    // set status of IOCTL
	    pGpioGPIORespIoctlBuf->Result =GPIOReg;
	    pGpioGPIORespIoctlBuf->Status = IOCTL_STATUS_SUCCESS;
	    BytesReturned = 0;
	    IoCmdStatus = GPIO_STATUS_CMD_SUCCESS; // was executed
	    IoCmdStatus = copy_to_user((char *)&pUserSpaceGPIOCmdIoctlArg->RespBuf,
		    (char *)pGpioGPIORespIoctlBuf, sizeof(GpioCmdIoctlReq));
	    put_user(IoCmdStatus,
		    &pUserSpaceGPIOCmdIoctlArg->RespBuf.Status);
	    put_user(BytesReturned,
		  &pUserSpaceGPIOCmdIoctlArg->ReturnedLen);
	}
	break;

	case IOCTL_WAIT_FOR_SHIFTG_INTR:
	{
	    // User space pointer
            GpioCmdArg *pUserSpaceGPIOCmdIoctlArg;
	    // Command request
            GpioCmdIoctlReq *pGpioGPIOCmdIoctlBuf;
	    // RespBuf pointer
	    GpioCmdIoctlReq *pGpioGPIORespIoctlBuf;

	    pUserSpaceGPIOCmdIoctlArg = (GpioCmdArg *) arg;

	    result = copy_from_user((char *)pGpioCmdArgBuf, 
		(char *)pUserSpaceGPIOCmdIoctlArg, sizeof(GpioCmdArg));
            if (result) 
	    {
		IoCmdStatus = result;
		break;
	    }

	    // CmdBuf pointer
	    pGpioGPIOCmdIoctlBuf = &pGpioCmdArgBuf->CmdBuf;
	
	    // RespBuf pointer
	    pGpioGPIORespIoctlBuf = &pGpioCmdArgBuf->RespBuf;

	    memmove((void *)pGpioGPIORespIoctlBuf, 
			(void *)pGpioGPIOCmdIoctlBuf, 
			sizeof(GpioCmdIoctlReq));
	    if ((IoCmdStatus = ProcessInterruptQueue(&shifty_intr_queue,
			&pGpioCard->shifty_controlLock_irqlock,
			&ShftgInterruptSemaphore,
			&GPIOReg)))
		goto get_out;

	    // set status of IOCTL
	    pGpioGPIORespIoctlBuf->Result = GPIOReg;
	    pGpioGPIORespIoctlBuf->Status = IOCTL_STATUS_SUCCESS;
	    BytesReturned = 0;
	    IoCmdStatus = GPIO_STATUS_CMD_SUCCESS; // was executed
	    IoCmdStatus = copy_to_user((char *)&pUserSpaceGPIOCmdIoctlArg->RespBuf,
		    (char *)pGpioGPIORespIoctlBuf, sizeof(GpioCmdIoctlReq));
	    put_user(IoCmdStatus,
		    &pUserSpaceGPIOCmdIoctlArg->RespBuf.Status);
	    put_user(BytesReturned,
		  &pUserSpaceGPIOCmdIoctlArg->ReturnedLen);
	}
	break;
	case IOCTL_WAIT_FOR_KVM_INTR:
	{
	    // User space pointer
            GpioCmdArg *pUserSpaceGPIOCmdIoctlArg;
	    // Command request
            GpioCmdIoctlReq *pGpioGPIOCmdIoctlBuf;
	    // RespBuf pointer
	    GpioCmdIoctlReq *pGpioGPIORespIoctlBuf;

	    pUserSpaceGPIOCmdIoctlArg = (GpioCmdArg *) arg;

	    result = copy_from_user((char *)pGpioCmdArgBuf, 
		(char *)pUserSpaceGPIOCmdIoctlArg, sizeof(GpioCmdArg));
            if (result) 
	    {
		IoCmdStatus = result;
		break;
	    }

	    // CmdBuf pointer
	    pGpioGPIOCmdIoctlBuf = &pGpioCmdArgBuf->CmdBuf;
	
	    // RespBuf pointer
	    pGpioGPIORespIoctlBuf = &pGpioCmdArgBuf->RespBuf;

	    memmove((void *)pGpioGPIORespIoctlBuf, 
			(void *)pGpioGPIOCmdIoctlBuf, 
			sizeof(GpioCmdIoctlReq));

	    // now wait for completion
	    if (pGpioCard->GPIOKVMInterruptInitialStatus != 0)
	    {
	    // If we have the status waiting when we loaded the
	    // driver, report the status immediately and don't
	    // wait.
		gpio_spin_lock(pGpioCard);
		GPIOReg = pGpioCard->GPIOKVMInterruptInitialStatus;
		pGpioCard->GPIOKVMInterruptInitialStatus = 0;
		gpio_spin_unlock(pGpioCard);
	    }
	    else
	    {
		if ((IoCmdStatus = ProcessInterruptQueue(&kvm_intr_queue,
			    &pGpioCard->kvm_controlLock_irqlock,
			    &KVMInterruptSemaphore,
			    &GPIOReg)))
		    goto get_out;
	    }

	    // set status of IOCTL
	    pGpioGPIORespIoctlBuf->Result = GPIOReg;
	    pGpioGPIORespIoctlBuf->Status = IOCTL_STATUS_SUCCESS;
	    BytesReturned = 0;
	    IoCmdStatus = GPIO_STATUS_CMD_SUCCESS; // was executed
	    IoCmdStatus = copy_to_user((char *)&pUserSpaceGPIOCmdIoctlArg->RespBuf,
		    (char *)pGpioGPIORespIoctlBuf, sizeof(GpioCmdIoctlReq));
	    put_user(IoCmdStatus,
		    &pUserSpaceGPIOCmdIoctlArg->RespBuf.Status);
	    put_user(BytesReturned,
		  &pUserSpaceGPIOCmdIoctlArg->ReturnedLen);
	}
	break;
	case IOCTL_WAIT_FOR_CP_INTR:
	{
	    // User space pointer
            GpioCmdArg *pUserSpaceGPIOCmdIoctlArg;
	    // Command request
            GpioCmdIoctlReq *pGpioGPIOCmdIoctlBuf;
	    // RespBuf pointer
	    GpioCmdIoctlReq *pGpioGPIORespIoctlBuf;

	    pUserSpaceGPIOCmdIoctlArg = (GpioCmdArg *) arg;

	    result = copy_from_user((char *)pGpioCmdArgBuf, 
		(char *)pUserSpaceGPIOCmdIoctlArg, sizeof(GpioCmdArg));
            if (result) 
	    {
		IoCmdStatus = result;
		break;
	    }

	    // CmdBuf pointer
	    pGpioGPIOCmdIoctlBuf = &pGpioCmdArgBuf->CmdBuf;
	
	    // RespBuf pointer
	    pGpioGPIORespIoctlBuf = &pGpioCmdArgBuf->RespBuf;

	    memmove((void *)pGpioGPIORespIoctlBuf, 
			(void *)pGpioGPIOCmdIoctlBuf, 
			sizeof(GpioCmdIoctlReq));
	    if ((IoCmdStatus = ProcessInterruptQueue(&cp_intr_queue,
			&pGpioCard->cp_controlLock_irqlock,
			&CPInterruptSemaphore,
			&GPIOReg)))
		goto get_out;

	    // set status of IOCTL
	    pGpioGPIORespIoctlBuf->Result = GPIOReg;
	    pGpioGPIORespIoctlBuf->Status = IOCTL_STATUS_SUCCESS;
	    BytesReturned = 0;
	    IoCmdStatus = GPIO_STATUS_CMD_SUCCESS; // was executed
	    IoCmdStatus = copy_to_user((char *)&pUserSpaceGPIOCmdIoctlArg->RespBuf,
		    (char *)pGpioGPIORespIoctlBuf, sizeof(GpioCmdIoctlReq));
	    put_user(IoCmdStatus,
		    &pUserSpaceGPIOCmdIoctlArg->RespBuf.Status);
	    put_user(BytesReturned,
		  &pUserSpaceGPIOCmdIoctlArg->ReturnedLen);
	}
	break;

	case IOCTL_WAIT_FOR_BLADE_INTR:
	{
	    // User space pointer
            GpioCmdArg *pUserSpaceGPIOCmdIoctlArg;
	    // Command request
            GpioCmdIoctlReq *pGpioGPIOCmdIoctlBuf;
	    // RespBuf pointer
	    GpioCmdIoctlReq *pGpioGPIORespIoctlBuf;

	    pUserSpaceGPIOCmdIoctlArg = (GpioCmdArg *) arg;

	    result = copy_from_user((char *)pGpioCmdArgBuf, 
		(char *)pUserSpaceGPIOCmdIoctlArg, sizeof(GpioCmdArg));
            if (result) 
	    {
		IoCmdStatus = result;
		break;
	    }

	    // CmdBuf pointer
	    pGpioGPIOCmdIoctlBuf = &pGpioCmdArgBuf->CmdBuf;
	
	    // RespBuf pointer
	    pGpioGPIORespIoctlBuf = &pGpioCmdArgBuf->RespBuf;

	    memmove((void *)pGpioGPIORespIoctlBuf, 
			(void *)pGpioGPIOCmdIoctlBuf, 
			sizeof(GpioCmdIoctlReq));
	    if ((IoCmdStatus = ProcessInterruptQueue(&blade_intr_queue,
			&pGpioCard->blade_controlLock_irqlock,
			&BladeInterruptSemaphore,
			&GPIOReg)))
		goto get_out;

	    // set status of IOCTL
	    pGpioGPIORespIoctlBuf->Result = GPIOReg;
	    pGpioGPIORespIoctlBuf->Status = IOCTL_STATUS_SUCCESS;
	    BytesReturned = 0;
	    IoCmdStatus = GPIO_STATUS_CMD_SUCCESS; // was executed
	    IoCmdStatus = copy_to_user((char *)&pUserSpaceGPIOCmdIoctlArg->RespBuf,
		    (char *)pGpioGPIORespIoctlBuf, sizeof(GpioCmdIoctlReq));
	    put_user(IoCmdStatus,
		    &pUserSpaceGPIOCmdIoctlArg->RespBuf.Status);
	    put_user(BytesReturned,
		  &pUserSpaceGPIOCmdIoctlArg->ReturnedLen);
	}
	break;
	case IOCTL_WAIT_FOR_IOM_INTR:
	{
	    // User space pointer
            GpioCmdArg *pUserSpaceGPIOCmdIoctlArg;
	    // Command request
            GpioCmdIoctlReq *pGpioGPIOCmdIoctlBuf;
	    // RespBuf pointer
	    GpioCmdIoctlReq *pGpioGPIORespIoctlBuf;

	    pUserSpaceGPIOCmdIoctlArg = (GpioCmdArg *) arg;

	    result = copy_from_user((char *)pGpioCmdArgBuf, 
		(char *)pUserSpaceGPIOCmdIoctlArg, sizeof(GpioCmdArg));
            if (result) 
	    {
		IoCmdStatus = result;
		break;
	    }

	    // CmdBuf pointer
	    pGpioGPIOCmdIoctlBuf = &pGpioCmdArgBuf->CmdBuf;
	
	    // RespBuf pointer
	    pGpioGPIORespIoctlBuf = &pGpioCmdArgBuf->RespBuf;

	    memmove((void *)pGpioGPIORespIoctlBuf, 
			(void *)pGpioGPIOCmdIoctlBuf, 
			sizeof(GpioCmdIoctlReq));

	    // now wait for completion
	    if (pGpioCard->GPIOIOMInterruptInitialStatus != 0)
	    {
	    // If we have the status waiting when we loaded the
	    // driver, report the status immediately and don't
	    // wait.
		gpio_spin_lock(pGpioCard);
		GPIOReg = pGpioCard->GPIOIOMInterruptInitialStatus;
		pGpioCard->GPIOIOMInterruptInitialStatus = 0;
		gpio_spin_unlock(pGpioCard);
	    }
	    else
	    {
		if ((IoCmdStatus = ProcessInterruptQueue(&iom_intr_queue,
			    &pGpioCard->iom_controlLock_irqlock,
			    &IOMInterruptSemaphore,
			    &GPIOReg)))
		    goto get_out;
	    } // end of if pGpioCard->GPIOIOMInterruptInitialStatus != 0

	    // set status of IOCTL
	    pGpioGPIORespIoctlBuf->Result = GPIOReg;
	    pGpioGPIORespIoctlBuf->Status = IOCTL_STATUS_SUCCESS;
	    BytesReturned = 0;
	    IoCmdStatus = GPIO_STATUS_CMD_SUCCESS; // was executed
	    IoCmdStatus = copy_to_user((char *)&pUserSpaceGPIOCmdIoctlArg->RespBuf,
		    (char *)pGpioGPIORespIoctlBuf, sizeof(GpioCmdIoctlReq));
	    put_user(IoCmdStatus,
		    &pUserSpaceGPIOCmdIoctlArg->RespBuf.Status);
	    put_user(BytesReturned,
		  &pUserSpaceGPIOCmdIoctlArg->ReturnedLen);
	}
	break;
	case IOCTL_WAIT_FOR_ECC_INTR:
	{
	    // User space pointer
	    GpioCmdArg *pUserSpaceGPIOCmdIoctlArg;
	    // Command request
	    GpioCmdIoctlReq *pGpioGPIOCmdIoctlBuf;
	    // RespBuf pointer
	    GpioCmdIoctlReq *pGpioGPIORespIoctlBuf;
	    unsigned long gpioflags;

	    pUserSpaceGPIOCmdIoctlArg = (GpioCmdArg *) arg;

	    result = copy_from_user((char *)pGpioCmdArgBuf, 
		(char *)pUserSpaceGPIOCmdIoctlArg, sizeof(GpioCmdArg));
	    if (result) 
	    {
		IoCmdStatus = result;
		break;
	    }

	    // CmdBuf pointer
	    pGpioGPIOCmdIoctlBuf = &pGpioCmdArgBuf->CmdBuf;
	
	    // RespBuf pointer
	    pGpioGPIORespIoctlBuf = &pGpioCmdArgBuf->RespBuf;

	    memmove((void *)pGpioGPIORespIoctlBuf, 
			(void *)pGpioGPIOCmdIoctlBuf, 
			sizeof(GpioCmdIoctlReq));
	    // now wait for completion
	    if (wait_event_interruptible (ECCInterruptSemaphore, (ECC_event_present != 0)))
	    {
		    IoCmdStatus = -ERESTARTSYS;
		    goto get_out;
	    }
	    if (driver_exited)
	    {
		    IoCmdStatus = -EINVAL;
		    goto get_out;
	    }

	    // set status of IOCTL
	    ECC_event_present = 0;
	    pGpioGPIORespIoctlBuf->Result = pGpioCard->GPIOECCInterruptStatus;
	    pGpioGPIORespIoctlBuf->Status = IOCTL_STATUS_SUCCESS;
	    BytesReturned = 0;
	    IoCmdStatus = GPIO_STATUS_CMD_SUCCESS; // was executed
	    IoCmdStatus = copy_to_user((char *)&pUserSpaceGPIOCmdIoctlArg->RespBuf,
		    (char *)pGpioGPIORespIoctlBuf, sizeof(GpioCmdIoctlReq));
	    put_user(IoCmdStatus,
		    &pUserSpaceGPIOCmdIoctlArg->RespBuf.Status);
	    put_user(BytesReturned,
		  &pUserSpaceGPIOCmdIoctlArg->ReturnedLen);
	}
	break;
	case IOCTL_WAIT_FOR_PSU_INTR:
	{
	    // User space pointer
	    GpioCmdArg *pUserSpaceGPIOCmdIoctlArg;
	    // Command request
	    GpioCmdIoctlReq *pGpioGPIOCmdIoctlBuf;
	    // RespBuf pointer
	    GpioCmdIoctlReq *pGpioGPIORespIoctlBuf;

	    pUserSpaceGPIOCmdIoctlArg = (GpioCmdArg *) arg;

	    result = copy_from_user((char *)pGpioCmdArgBuf, 
		(char *)pUserSpaceGPIOCmdIoctlArg, sizeof(GpioCmdArg));
	    if (result) 
	    {
		IoCmdStatus = result;
		break;
	    }

	    // CmdBuf pointer
	    pGpioGPIOCmdIoctlBuf = &pGpioCmdArgBuf->CmdBuf;
	
	    // RespBuf pointer
	    pGpioGPIORespIoctlBuf = &pGpioCmdArgBuf->RespBuf;

	    memmove((void *)pGpioGPIORespIoctlBuf, 
			(void *)pGpioGPIOCmdIoctlBuf, 
			sizeof(GpioCmdIoctlReq));

	    // now wait for completion
	    if (pGpioCard->GPIOPSUInterruptInitialStatus != 0)
	    {
	    // If we have the status waiting when we loaded the
	    // driver, report the status immediately and don't
	    // wait.
		gpio_spin_lock(pGpioCard);
		GPIOReg = pGpioCard->GPIOPSUInterruptInitialStatus;
		pGpioGPIORespIoctlBuf->Result =GPIOReg;
		pGpioCard->GPIOPSUInterruptInitialStatus = 0;
		gpio_spin_unlock(pGpioCard);
	    }
	    else
	    {
		if ((IoCmdStatus = ProcessInterruptQueue(&psu_intr_queue,
			    &pGpioCard->psu_controlLock_irqlock,
			    &PSUInterruptSemaphore,
			    &GPIOReg)))
		    goto get_out;
	    }

	    // set status of IOCTL
	    pGpioGPIORespIoctlBuf->Result =GPIOReg;
	    pGpioGPIORespIoctlBuf->Status = IOCTL_STATUS_SUCCESS;
	    BytesReturned = 0;
	    IoCmdStatus = GPIO_STATUS_CMD_SUCCESS; // was executed
	    IoCmdStatus = copy_to_user((char *)&pUserSpaceGPIOCmdIoctlArg->RespBuf,
		    (char *)pGpioGPIORespIoctlBuf, sizeof(GpioCmdIoctlReq));
	    put_user(IoCmdStatus,
		    &pUserSpaceGPIOCmdIoctlArg->RespBuf.Status);
	    put_user(BytesReturned,
		  &pUserSpaceGPIOCmdIoctlArg->ReturnedLen);
	}
	break;

	case IOCTL_SIMULATE_GENERATE_BLADE_INTR:
	{
	    // Simulate the status for testing purpose
	    pGpioCard->GPIOBladeInterruptStatus = 0x1000; 
	    status = QueueISRStatus(&blade_intr_queue,
		    pGpioCard->GPIOBladeInterruptStatus,
		    &BladeInterruptSemaphore);
	}
	break;

	case IOCTL_SIMULATE_GENERATE_FAN_INTR:
	{
	    // Simulate the status for testing purpose
	    pGpioCard->GPIOFanInterruptStatus = 0x1100; 
	    status = QueueISRStatus(&fan_intr_queue,
		    pGpioCard->GPIOFanInterruptStatus,
		    &FanInterruptSemaphore);
	}
	break;

	case IOCTL_SIMULATE_GENERATE_PSU_INTR:
	{
	    // Simulate the status for testing purpose
	    pGpioCard->GPIOPSUInterruptStatus = 0x1010; 
	    status = QueueISRStatus(&psu_intr_queue,
		    pGpioCard->GPIOPSUInterruptStatus,
		    &PSUInterruptSemaphore);
	}
	break;


	case IOCTL_SIMULATE_GENERATE_KVM_INTR:
	{
		int i;

		for (i=0; i < 128; i++) {
			static u16 val=0;

			// Simulate the status for testing purpose
			pGpioCard->GPIOKVMInterruptStatus = val++; 
			status = QueueISRStatus(&kvm_intr_queue,
				pGpioCard->GPIOKVMInterruptStatus,
				&KVMInterruptSemaphore);
		}
	}
	break;

	case IOCTL_SIMULATE_GENERATE_CP_INTR:
	{
	    // Simulate the status for testing purpose
	    pGpioCard->GPIOCPInterruptStatus = 0x1110; 
	    status = QueueISRStatus(&cp_intr_queue,
		    pGpioCard->GPIOCPInterruptStatus,
		    &CPInterruptSemaphore);
	}
	break;

	case IOCTL_SIMULATE_GENERATE_IOM_INTR:
	{
	    // Simulate the status for testing purpose
	    pGpioCard->GPIOIOMInterruptStatus = 0x1001; 
	    status = QueueISRStatus(&iom_intr_queue,
		    pGpioCard->GPIOIOMInterruptStatus,
		    &IOMInterruptSemaphore);
	}
	break;

	case IOCTL_SIMULATE_GENERATE_ECC_INTR:
	{

			// Simulate the status for testing purpose
			pGpioCard->GPIOECCInterruptStatus = 0x1101; 
			ECC_event_present = 1;
			wake_up_interruptible(&ECCInterruptSemaphore);
	}
	break;

    	case IOCTL_RESET_XILINX_IOM1:
    	{
		xilinx_recovery_event_present = 1;
		// Set up the bus index so the recovery thread knows
		// this is for PPC bus 0
		pGpioCard->PPCBusIndex = 0;
		wake_up(&XilinxRecoverySemaphore);
	}
	break;

    	case IOCTL_RESET_XILINX_IOM2:
    	{
		xilinx_recovery_event_present = 1;
		// Set up the bus index so the recovery thread knows
		// this is for PPC bus 0
		pGpioCard->PPCBusIndex = 1;
		wake_up(&XilinxRecoverySemaphore);
	}
	break;
	
    	case IOCTL_DISABLE_XILINX_IOM1:
    	{
		xilinx_disable_event_present = 1;
		// Set up the bus index so the recovery thread knows
		// this is for PPC bus 0
		pGpioCard->PPCBusIndex = 0;
		wake_up(&XilinxDisableSemaphore);
	}
	break;

    	case IOCTL_DISABLE_XILINX_IOM2:
    	{
		xilinx_disable_event_present = 1;
		// Set up the bus index so the recovery thread knows
		// this is for PPC bus 0
		pGpioCard->PPCBusIndex = 1;
		wake_up(&XilinxDisableSemaphore);
	}
	break;

    	case IOCTL_RESET_XILINX_PSU1:
    	{
		xilinx_recovery_event_present = 1;
		// Set up the bus index so the recovery thread knows
		// this is for PPC bus 0
		pGpioCard->PPCBusIndex = 2;
		wake_up(&XilinxRecoverySemaphore);
	}
	break;

    	case IOCTL_RESET_XILINX_PSU2:
    	{
		xilinx_recovery_event_present = 1;
		// Set up the bus index so the recovery thread knows
		// this is for PPC bus 0
		pGpioCard->PPCBusIndex = 3;
		wake_up(&XilinxRecoverySemaphore);
	}
	break;

    	case IOCTL_RESET_XILINX_KVM:
    	{
		xilinx_recovery_event_present = 1;
		// Set up the bus index so the recovery thread knows
		// this is for PPC bus 0
		pGpioCard->PPCBusIndex = 4;
		wake_up(&XilinxRecoverySemaphore);
	}
	break;

    	case IOCTL_RESET_PPC_I2C_0:
    	{
		xilinx_recovery_event_present = 1;
		// Set up the bus index so the recovery thread knows
		// this is for PPC bus 0
		pGpioCard->PPCBusIndex = 5;
		wake_up(&XilinxRecoverySemaphore);
	}
	break;

    	case IOCTL_RESET_PPC_I2C_1:
    	{
		xilinx_recovery_event_present = 1;
		// Set up the bus index so the recovery thread knows
		// this is for PPC bus 0
		pGpioCard->PPCBusIndex = 6;
		wake_up(&XilinxRecoverySemaphore);
	}
	break;

	case IOCTL_SIMULATE_GENERATE_SHIFTY_INTR:
	{
	    // Simulate the status for testing purpose
	    pGpioCard->GPIOShiftyInterruptStatus = 0x1111; 
	    status = QueueISRStatus(&shifty_intr_queue,
		    pGpioCard->GPIOShiftyInterruptStatus,
		    &ShftgInterruptSemaphore);
	}
	break;

	case IOCTL_ENABLE_KVM_INTR:
	{
		enable_irq(pGpioCard->PPCKVMIrqLevel);
	}
	break;
	case IOCTL_ENABLE_PSU_INTR:
	{
		enable_irq(pGpioCard->PPCPSUIrqLevel);
	}
	break;
	case IOCTL_ENABLE_IOM_INTR:
	{
		enable_irq(pGpioCard->PPCIOMIrqLevel);
	}
	break;
	case IOCTL_ENABLE_FAN_INTR:
	{
		enable_irq(pGpioCard->PPCFanIrqLevel);
	}
	break;
	case IOCTL_ENABLE_SHIFTY_INTR:
	{
		enable_irq(pGpioCard->PPCShftgIrqLevel);
	}
	break;
	case IOCTL_ENABLE_BLADE_INTR:
	{
		enable_irq(pGpioCard->PPCBladeIrqLevel);
	}
	break;
	case IOCTL_ENABLE_CP_INTR:
	{
		enable_irq(pGpioCard->PPCCPIrqLevel);
	}
	break;

	case IOCTL_DISABLE_KVM_INTR:
	{
		disable_irq(pGpioCard->PPCKVMIrqLevel);
	}
	break;
	case IOCTL_DISABLE_PSU_INTR:
	{
		disable_irq(pGpioCard->PPCPSUIrqLevel);
	}
	break;
	case IOCTL_DISABLE_IOM_INTR:
	{
		disable_irq(pGpioCard->PPCIOMIrqLevel);
	}
	break;
	case IOCTL_DISABLE_FAN_INTR:
	{
		disable_irq(pGpioCard->PPCFanIrqLevel);
	}
	break;
	case IOCTL_DISABLE_SHIFTY_INTR:
	{
		disable_irq(pGpioCard->PPCShftgIrqLevel);
	}
	break;
	case IOCTL_DISABLE_BLADE_INTR:
	{
		disable_irq(pGpioCard->PPCBladeIrqLevel);
	}
	break;
	case IOCTL_DISABLE_CP_INTR:
	{
		disable_irq(pGpioCard->PPCCPIrqLevel);
	}
	break;
	case IOCTL_READ_SMBUS_TIMEOUT_STAT:
	{
	    // User space pointer
            GpioCmdArg *pUserSpaceGPIOCmdIoctlArg;
	    // Command request
            GpioCmdIoctlReq *pGpioGPIOCmdIoctlBuf;
	    // RespBuf pointer
	    GpioCmdIoctlReq *pGpioGPIORespIoctlBuf;


	    pUserSpaceGPIOCmdIoctlArg = (GpioCmdArg *) arg;

	    result = copy_from_user((char *)pGpioCmdArgBuf, 
		(char *)pUserSpaceGPIOCmdIoctlArg, sizeof(GpioCmdArg));
            if (result) 
	    {
		IoCmdStatus = result;
		break;
	    }

	    // CmdBuf pointer
	    pGpioGPIOCmdIoctlBuf = &pGpioCmdArgBuf->CmdBuf;
	
	    // RespBuf pointer
	    pGpioGPIORespIoctlBuf = &pGpioCmdArgBuf->RespBuf;

	    memmove((void *)pGpioGPIORespIoctlBuf, 
			(void *)pGpioGPIOCmdIoctlBuf, 
			sizeof(GpioCmdIoctlReq));

		gpio_spin_lock(pGpioCard);
		// Do GPIO stuff
		switch (pGpioCmdArgBuf->CmdBuf.ReqType)
		{
			case 0:
			// IOM 1
				pGpioGPIORespIoctlBuf->Result = pGpioCard->GPIOIOM1SMBUSTimeoutCount;
			break;
			case 1:
			// IOM 2
				pGpioGPIORespIoctlBuf->Result = pGpioCard->GPIOIOM2SMBUSTimeoutCount;
			break;
			case 2:
			// PSU 1
				pGpioGPIORespIoctlBuf->Result = pGpioCard->GPIOPSU1SMBUSTimeoutCount;
			break;
			case 3:
			// PSU 2
				pGpioGPIORespIoctlBuf->Result = pGpioCard->GPIOPSU2SMBUSTimeoutCount;
			break;
			case 4:
			// KVM
				pGpioGPIORespIoctlBuf->Result = pGpioCard->GPIOKVMSMBUSTimeoutCount;
			break;
		}
		gpio_spin_unlock(pGpioCard);

		// set status of IOCTL
		pGpioGPIORespIoctlBuf->Status = IOCTL_STATUS_SUCCESS;
		BytesReturned = 0;
		IoCmdStatus = GPIO_STATUS_CMD_SUCCESS; // was executed
		IoCmdStatus = copy_to_user((char *)&pUserSpaceGPIOCmdIoctlArg->RespBuf,
		 	(char *)pGpioGPIORespIoctlBuf, sizeof(GpioCmdIoctlReq));
	    	put_user(IoCmdStatus,
            	  	&pUserSpaceGPIOCmdIoctlArg->RespBuf.Status);
		put_user(BytesReturned,
           	      &pUserSpaceGPIOCmdIoctlArg->ReturnedLen);
	}
#if 0
	case IOCTL_TESTING_CORE_DUMP:
	{
		// testing purpose only, will never get called
		char *bad_buff;
		// Will panic the system
		bad_buff = (char *)kmalloc(256, GFP_KERNEL);	
		printk("************* Panicking the system **********\n");
		memcpy(0x0000, bad_buff, 256);
		kfree (bad_buff);
	}
	break;
#endif
	case IOCTL_GET_BLADE_INTR_QUEUE:
	case IOCTL_GET_FAN_INTR_QUEUE:
	case IOCTL_GET_IOM_INTR_QUEUE:
	case IOCTL_GET_PSU_INTR_QUEUE:
	case IOCTL_GET_CP_INTR_QUEUE:
	case IOCTL_GET_KVM_INTR_QUEUE:
	case IOCTL_GET_SHIFTY_INTR_QUEUE:
	{
	    struct intr_status_t *intr_status;

	    switch (cmd_in) {
	    case IOCTL_GET_BLADE_INTR_QUEUE:
		intr_status = &blade_intr_queue;
		break;
	    case IOCTL_GET_FAN_INTR_QUEUE:
		intr_status = &fan_intr_queue;
		break;
	    case IOCTL_GET_IOM_INTR_QUEUE:
		intr_status = &iom_intr_queue;
		break;
	    case IOCTL_GET_PSU_INTR_QUEUE:
		intr_status = &psu_intr_queue;
		break;
	    case IOCTL_GET_CP_INTR_QUEUE:
		intr_status = &cp_intr_queue;
		break;
	    case IOCTL_GET_KVM_INTR_QUEUE:
		intr_status = &kvm_intr_queue;
		break;
	    case IOCTL_GET_SHIFTY_INTR_QUEUE:
		intr_status = &shifty_intr_queue;
		break;
	    }
	    IoCmdStatus = copy_to_user((char *)arg, (char *)intr_status, sizeof(*intr_status));
	}

	default:
	{	
	}
	break;
    }
get_out:
	kfree (pGpioCmdArgBuf);
   	return IoCmdStatus;
}

int ProcessInterruptQueue(struct intr_status_t *intr_queue,
		spinlock_t *irqlock,
		wait_queue_head_t *InterruptSemaphore,
		u32 *GPIOReg)
{
	unsigned long gpioflags;
	int status=0;

	// IMPORTANT: lock must be held during a) find EMPTYFIFO and b) set waiting
	spin_lock_irqsave(*irqlock, gpioflags);
	if (ISREMPTYFIFO(intr_queue)) {
		if (intr_queue->waiting)
			printk("hapi: %s already waiting\n", __FUNCTION__);
		intr_queue->waiting = 1;
		spin_unlock_irqrestore(*irqlock, gpioflags);
		if (wait_event_interruptible(*InterruptSemaphore, !intr_queue->waiting)) {
			status = -ERESTARTSYS;
			goto get_out;
		}
		if (driver_exited) {
			status = -EINVAL;
			goto get_out;
		}
		spin_lock_irqsave(*irqlock, gpioflags);
		if (ISREMPTYFIFO(intr_queue)) {
			printk("hapi: %s queue empty\n", __FUNCTION__);
			status = -1;
			spin_unlock_irqrestore(*irqlock, gpioflags);
			goto get_out;
		}
	}

	//printk("ProcessInterruptQueue, *GPIOReg: 0x%x.\n", *GPIOReg);
	*GPIOReg = intr_queue->fifo[intr_queue->out];
	intr_queue->outcnt++;

        /* increment and wrap pointer, if necessary */
        if (++intr_queue->out >= intr_queue->fifosize)
                intr_queue->out = 0;
	spin_unlock_irqrestore(*irqlock, gpioflags);
get_out:
	return status;
}
